/* Output from p2c, the Pascal-to-C translator */
/* From input file "fnn.mod" */


#include <p2c/p2c.h>


/*
 *
 *      FALSE NEAREST NEIGHBORS CODE
 *
 *
 *      Usage:
 *
 *      fnn -d lowdim highdim -N Npoints -t time-delayembedding
 *          -B datafile -r RTOL
 *
 */

#ifndef DynArray_H
#include "DynArray.h"
#endif

#ifndef DynMatrix_H
#include "DynMatrix.h"
#endif

#ifndef Util_H
#include "Util.h"
#endif

#ifndef KdTree_H
#include "KdTree.h"
#endif


#ifndef IO_H
#include "IO.h"
#endif



Static double *OneDData;   /* The 1-D data array */
Static boolean DoLPS;   /* do we run LPS? */
Static unsigned long Ndata;   /* the number of data points */

Static unsigned long Nvectors;
/* The number of vectors that we can
                                        form given the dimension and time
                                        delay */

Static unsigned long DimLo, DimHi;   /* range of dimensions to loop over */
Static unsigned long Tau;
/* time delay, in units of sampling
                                        time*/

Static double Rtol;

Static double Size;   /* typical size of the attractor */

Static unsigned long d;


Static void fnn_ReadInData(void)
{
    /*
     *    From the arguments passed to the program get the input data vector
     *    Either from an ascii file, or from a binary file:
     *
     *    fnn -A filename.ascii
     *   or
     *    fnn -B filename.binary
     *
     *    -B is searched for FIRST on the command line
     *    Allocates OneDData;
     */
    unsigned long argplace;
    FILE *binfile;
    FILE *asciifile;
    unsigned long i;
    Char TEMP[256];
    Char STR1[256];
    unsigned long FORLIM;

    asciifile = NULL;
    binfile = NULL;
    argplace = Util_ArgvStringExists("-N");
    if (argplace == 0)
	Util_FatalError("You must provide a -N npoints flag!");

    strcpy(TEMP, P_argv[argplace + 1]);
    Ndata = Util_ScanInteger(TEMP);
    printf("Ndata = %12lu\n", Ndata);

    OneDData = DynArray_MakeDoubleArray(Ndata);
    /*allocate the array */


    argplace = Util_ArgvStringExists("-B");
    if (argplace > 0) {   /* found binary file */
	if (binfile != NULL)
	    binfile = freopen(strcpy(STR1, P_argv[argplace + 1]), "rb",
			      binfile);
	else
	    binfile = fopen(strcpy(STR1, P_argv[argplace + 1]), "rb");
	if (binfile == NULL)
	    _EscIO(FileNotFound);
	fread(OneDData, sizeof(double) * Ndata, 1, binfile);
    } else {
	argplace = Util_ArgvStringExists("-A");
	if (argplace > 0) {
	    if (asciifile != NULL)
		asciifile = freopen(strcpy(STR1, P_argv[argplace + 1]), "r",
				    asciifile);
	    else
		asciifile = fopen(strcpy(STR1, P_argv[argplace + 1]), "r");
	    if (asciifile == NULL)
		_EscIO(FileNotFound);
	    FORLIM = Ndata;
	    for (i = 0; i < FORLIM; i++)
		fscanf(asciifile, "%lg", &OneDData[i]);
	} else
	    Util_FatalError(
	      "You must provide either a -A (ascii) or -B (binary) file");
    }

    if (binfile != NULL)
	fclose(binfile);
    if (asciifile != NULL)
	fclose(asciifile);
}


Static void fnn_GetOptions(void)
{
    /*
     *  Get the other pieces of data from the command line :
     *
     *  -d dim1 dim2
     */
    unsigned long argplace;
    Char TEMP[256];

    if (P_argc == 1) {
	IO_WritelnS(&_iob[2], 1, 6, "usage:");
	IO_WritelnS(&_iob[2], 1, 42,
		    "fnn -N npoints -d dim1 dim2 -t tau -r rtol");
	IO_WritelnS(&_iob[2], 1, 45,
		    "{-LPS} {-B datafile.dbin}|{-A datafile.ascii}");

	exit(0);
    }

    argplace = Util_ArgvStringExists("-d");
    if (argplace == 0)
	Util_FatalError("You must provide a -d flag!!!");
    strcpy(TEMP, P_argv[argplace + 1]);
    DimLo = Util_ScanInteger(TEMP);
    strcpy(TEMP, P_argv[argplace + 2]);
    DimHi = Util_ScanInteger(TEMP);

    assert(DimLo < DimHi);

    argplace = Util_ArgvStringExists("-t");
    if (argplace == 0)
	Util_FatalError("You must provide a -t flag!!!");
    strcpy(TEMP, P_argv[argplace + 1]);
    Tau = Util_ScanInteger(TEMP);

    argplace = Util_ArgvStringExists("-r");
    if (argplace == 0)
	Util_FatalError("You must provide a -r flag!!!");
    strcpy(TEMP, P_argv[argplace + 1]);
    Rtol = Util_ScanLongReal(TEMP);

    argplace = Util_ArgvStringExists("-LPS");
    if (argplace == 0)
	DoLPS = false;
    else
	DoLPS = true;
    printf("DimLo:%5lu DimHi:%5lu Tau:%12lu Rtol:%6.2f\n",
	   DimLo, DimHi, Tau, Rtol);


}


Static float **fnn_EmbedData(double *oned, unsigned long ndata,
  unsigned long d, unsigned long tau, unsigned long *nvec)
{
    /*
     *  Do the time-delay embedding, dynamically creating a "RealMatrixPtr",
     *  size Length nvec, width d.
     *  returning the number of valid vectors into nvec.
     *  the RealMatrixPtr is accessed as      m^[0..d-1]^[0..nvec-1]
     */
    unsigned long i, j;
    float **m;
    unsigned long FORLIM;


    *nvec = ndata - d * tau;
    m = DynMatrix_MakeMatrix(*nvec, d);

    FORLIM = *nvec;
    for (i = 0; i < FORLIM; i++) {
	for (j = 0; j < d; j++)
	    m[i][j] = oned[i + j * tau];
    }
    return m;

}


Static void fnn_DoSizeOfAttractor(void)
{
    unsigned long i;
    double m, d;
    unsigned long FORLIM;
    double TEMP;


    m = 0.0;
    FORLIM = Ndata;
    for (i = 0; i < FORLIM; i++)
	m += OneDData[i];
    m /= Ndata;

    d = 0.0;
    FORLIM = Ndata;
    for (i = 0; i < FORLIM; i++) {
	TEMP = OneDData[i] - m;
	d += TEMP * TEMP;
    }

    Size = sqrt(d / Ndata);

}



Static void fnn_CalculateFalseNeighbors(unsigned long d)
{
    /*
     * actually calculate the false neighbors.
     */
    float **embeddata;
    tTopRec *tree;
    unsigned long i, nfalse;
    unsigned long *indexes;
    float *distances;
    double r1, rext;
    unsigned long FORLIM;

    printf("Embedding data, for d = %4lu\n", d);
    embeddata = fnn_EmbedData(OneDData, Ndata, d, Tau, &Nvectors);

    printf("Creating Tree: Nvectors = %12lu d = %12lu\n", Nvectors, d);
    tree = KdTree_CreateKdTree(embeddata, Nvectors, d);

    Nvectors--;

    indexes = DynArray_MakeCardArray(10);
    distances = DynArray_MakeRealArray(10);

    nfalse = 0;
    FORLIM = Nvectors;
    for (i = 0; i < FORLIM; i++) {
	KdTree_SearchKdTree(tree, embeddata[i], 2, indexes, distances);
	/* writeln('Neighbor index = ',indexes^[1]:6,'at dis ',distances^[1]:8); */

	r1 = sqrt(distances[1]);
	rext = OneDData[i + d * Tau] - OneDData[indexes[1] + d * Tau];
	if (rext < 0)
	    rext = -rext;
	/* rext is the extra distance incurred for going to the next highest dimension */

	if (rext / r1 > Rtol)   /* we have a false neighbor */
	    nfalse++;
	else if (rext / Size > 2.0)
	    nfalse++;

	if (i % 100 == 0) {
	    putchar('.');
	    fflush(stdout);
	}

    }
    /* find the nearest neighbor! */


    printf("\nProportion of f.n.n. = %6lu/%6lu%10.3f\n",
	   nfalse, Nvectors, (double)nfalse / Nvectors);

    printf("Destroying Tree...\n");
    KdTree_DestroyKdTree(&tree);
    DynMatrix_FreeMatrix(&embeddata, Nvectors, d);


    /* second test */
}


Static double fnn_EuclideanDistance(unsigned long index1,
				    unsigned long index2, unsigned long d)
{
    double dis, tmp;
    unsigned long i;

    dis = 0.0;
    for (i = 0; i <= d; i++) {
	tmp = OneDData[index1 + i * Tau] - OneDData[index2 + i * Tau];
	dis += tmp * tmp;
    }

    return sqrt(dis);

}


Static void fnn_TwoEuclideanDistances(unsigned long index1,
  unsigned long index2, unsigned long d, double *outd, double *outdp1)
{
    double dis, tmp;
    unsigned long i;

    dis = 0.0;
    for (i = 0; i < d; i++) {
	tmp = OneDData[index1 + i * Tau] - OneDData[index2 + i * Tau];
	dis += tmp * tmp;
    }

    *outd = sqrt(dis);

    tmp = OneDData[index1 + d * Tau] - OneDData[index2 + d * Tau];
    dis += tmp * tmp;

    *outdp1 = sqrt(dis);

}


#define nneigh          1


Static void fnn_CalculateLPS(unsigned long d)
{
    /*
     * calculate the LPS statistic.
     */
    float **embeddata1, **embeddata2;
    tTopRec *tree1, *tree2;
    unsigned long i, j, nfalse;
    unsigned long *indexes1, *indexes2;
    float *distances1, *distances2;
    double a1, a2, a3, a4, Q1, Q2, W, W2;
    unsigned long FORLIM;

    printf("Embedding data, for d = %4lu\n", d);
    embeddata1 = fnn_EmbedData(OneDData, Ndata, d, Tau, &Nvectors);

    printf("Creating First Tree...\n");
    tree1 = KdTree_CreateKdTree(embeddata1, Nvectors, d);

    printf("Embedding data, for d = %4lu\n", d + 1);
    embeddata2 = fnn_EmbedData(OneDData, Ndata, d + 1, Tau, &Nvectors);

    printf("Creating Second Tree...\n");
    tree2 = KdTree_CreateKdTree(embeddata2, Nvectors, d + 1);



    Nvectors--;

    indexes1 = DynArray_MakeCardArray(20);
    distances1 = DynArray_MakeRealArray(20);
    indexes2 = DynArray_MakeCardArray(20);
    distances2 = DynArray_MakeRealArray(20);

    nfalse = 0;

    W = 0.0;

    FORLIM = Nvectors;
    for (i = 0; i < FORLIM; i++) {
	KdTree_SearchKdTree(tree1, embeddata1[i], nneigh + 1, indexes1,
			    distances1);
	KdTree_SearchKdTree(tree2, embeddata2[i], nneigh + 1, indexes2,
			    distances2);
	/* writeln('Neighbor index = ',indexes^[1]:6,'at dis ',distances^[1]:8); */

	W2 = 1.0;
	for (j = 1; j <= nneigh; j++) {
	    /* Calculate distances */
	    /*    Q1 := EuclideanDistance(i, indexes1^[j], d+1) /
	                      EuclideanDistance(i, indexes2^[j], d+1);

	                Q2 := EuclideanDistance(i, indexes1^[j], d)/
	                      EuclideanDistance(i, indexes2^[j], d);
	     */

	    fnn_TwoEuclideanDistances(i, indexes1[j], d, &a3, &a1);
	    fnn_TwoEuclideanDistances(i, indexes2[j], d, &a4, &a2);
	    Q1 = a1 / a2;
	    Q2 = a3 / a4;


	    W2 *= Q1 * Q2;
	}


	/*writeln('Local W = ',EXP( LN(W2) / (2*nneigh) ) ); */
	W += exp(log(W2) / (nneigh * 2));
	/* take the geometric mean */

	if (i % 100 == 0) {
	    putchar('.');
	    fflush(stdout);
	}

    }
    /* find the nearest neighbor! */



    W = log(W / Nvectors);

    printf("\nd, W, W/T = %5lu % .3E % .3E\n", d, W, W / Tau);

    printf("Destroying Tree...\n");
    KdTree_DestroyKdTree(&tree1);
    DynMatrix_FreeMatrix(&embeddata1, Nvectors, d);

    KdTree_DestroyKdTree(&tree2);
    DynMatrix_FreeMatrix(&embeddata2, Nvectors, d + 1);

}

#undef nneigh


main(int argc, Char *argv[])
{  /* the main program */
    unsigned long FORLIM;

    PASCAL_MAIN(argc, argv);
    fnn_GetOptions();
    fnn_ReadInData();
    fnn_DoSizeOfAttractor();

    FORLIM = DimHi;
    for (d = DimLo; d <= FORLIM; d++) {
	fnn_CalculateFalseNeighbors(d);
	if (DoLPS)
	    fnn_CalculateLPS(d);

    }


    exit(EXIT_SUCCESS);
}



/* End. */
